﻿using Algorand.V2.Algod.Model;
using Algorand.V2.Indexer.Model;
using AlgorandStakingBot.Model;
using AlgorandStakingBot.Utils;
using NLog;
using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace AlgorandStakingBot
{
    public class StakingMonitor
    {
        protected readonly StakingConfiguration configuration;
        protected readonly IndexerConfiguration indexerConfiguration;
        protected readonly Algorand.V2.Algod.DefaultApi algodClient;
        protected readonly Algorand.V2.Indexer.SearchApi searchApi;
        protected readonly Algorand.V2.Indexer.LookupApi lookupApi;
        protected readonly CancellationToken cancellationToken;
        protected readonly Logger logger = LogManager.GetCurrentClassLogger();
        protected ulong? LastInterval = null;
        protected ConcurrentBag<string> KnownLogicSigAccounts = new ConcurrentBag<string>();
        protected ConcurrentBag<string> KnownNonLogicSigAccounts = new ConcurrentBag<string>();

        public StakingMonitor(Configuration configuration, CancellationToken cancellationToken)
        {
            this.configuration = configuration.Staking;
            this.indexerConfiguration = configuration.Indexer;
            this.algodClient = AlgoExtensions.GetAlgod(configuration.Algod);
            this.searchApi = AlgoExtensions.GetSearchApi(configuration.Indexer);
            this.lookupApi = AlgoExtensions.GetLookupApi(configuration.Indexer);
            this.cancellationToken = cancellationToken;
        }


        public async Task Run()
        {
            var currentTime = (ulong)DateTimeOffset.Now.ToUnixTimeSeconds();


            LastInterval = currentTime / configuration.Interval;

            while (!cancellationToken.IsCancellationRequested)
            {
                var current = currentTime / configuration.Interval;
                while (LastInterval >= current)
                {
                    Console.Write(".");
                    await Task.Delay(1000, cancellationToken);
                    currentTime = (ulong)DateTimeOffset.Now.ToUnixTimeSeconds();
                    current = currentTime / configuration.Interval;
                }

                var algoParams = await algodClient.ParamsAsync();
                await ProcessNewStakingRound(algoParams);
                LastInterval = current;
            }

        }

        public async Task ProcessNewStakingRound(TransactionParametersResponse algoParams)
        {
            try
            {
                var round = algoParams.LastRound;
                logger.Info($"{DateTimeOffset.Now} Starting dispercing round {round}");
                var balances = await GetBalances(round);
                var balances2 = await GetBalances2(round);
                logger.Info($"{DateTimeOffset.Now} balances2: {balances2.Count()}");

                var resultList = balances.Where(n => balances2.Select(n1 => n1.Address).Contains(n.Address));

               var accounts = resultList
                                    .Where(b => !b.IsFrozen)
                                    .Select(b => b.Address)
                                    .Where(a => !configuration.ExcludedAccounts.Contains(a))
                                    .ToHashSet();
       logger.Info($"{DateTimeOffset.Now} balances: {accounts.Count()}");
                var toSend = resultList

                                    .Where(a => !configuration.ExcludedAccounts.Contains(a.Address));
                var toSendAmount = toSend.Sum(a => (long)a.Amount);
                var rewards = CalculateAccountReward(toSend);
                var rewardsAmount = rewards.Sum(r => Convert.ToDecimal(r.Value));
                  var keys = rewards.Keys.ToList();
#if DEBUG
                foreach (var r in rewards)
                {
                    logger.Info($"Reward:{r.Key}:{r.Value}");
                }
#endif
                for (var page = 0; page <= keys.Count / 20; page++)
                {
                    var pageRewards = rewards.Skip(page * 20).Take(20);

#if DEBUG
                    foreach (var r in pageRewards)
                    {
                        logger.Info($"Page:{page}:{r.Key}:{r.Value}");
                    }
#endif

                    var batch = PrepareBatch(pageRewards, algoParams);
                    logger.Info($"{DateTimeOffset.Now} {page} ToSend: {rewards.Count()}, batch {batch.Count()} accountsBalance {toSendAmount} rewards {rewardsAmount}");
                    var sent = await AlgoExtensions.SubmitTransactions(algodClient, batch);
                    logger.Info($"{DateTimeOffset.Now} {page} Sent: {sent.TxId}");

                }
        
            
             }
            catch (Algorand.V2.Algod.Model.ApiException<ErrorResponse> ex)
            {
                logger.Error(ex);
                logger.Error(ex.Result.Message);
                // If there is error we consider the account as logicsig
            }
            catch (Exception ex)
            {
                logger.Error(ex);
            }
        } 
     
        protected async Task<List<MiniAssetHolding>> GetBalances(ulong round)
        {
            string? next = null;
            var balances = new List<MiniAssetHolding>();
            await Task.Delay(indexerConfiguration.DelayMs, cancellationToken);
            var balance = await lookupApi.BalancesAsync(include_all: false, limit: 1000, next: next, round: round, currency_greater_than: configuration.MinimumBalanceForStaking, currency_less_than: null, asset_id: (int)configuration.AssetId, cancellationToken);
            balances.AddRange(balance.Balances);
            while (balance.Balances.Count > 0)
            {
                await Task.Delay(indexerConfiguration.DelayMs, cancellationToken);
                balance = await lookupApi.BalancesAsync(include_all: false, limit: 1000, next: balance.NextToken, round: round, currency_greater_than: configuration.MinimumBalanceForStaking, currency_less_than: null, asset_id: (int)configuration.AssetId, cancellationToken);
                balances.AddRange(balance.Balances);

            }
            return balances;
        }

    protected async Task<List<MiniAssetHolding>> GetBalances2(ulong round2)
        {
            string? next = null;
            var balances2 = new List<MiniAssetHolding>();
            await Task.Delay(indexerConfiguration.DelayMs, cancellationToken);
            var balance2 = await lookupApi.BalancesAsync(include_all: false, limit: 1000, next: next, round: round2, currency_greater_than: configuration.MinimumBalanceForStaking, currency_less_than: null, asset_id: (int)configuration.assetsend, cancellationToken);

            balances2.AddRange(balance2.Balances);
            while (balance2.Balances.Count > 0)
            {
                await Task.Delay(indexerConfiguration.DelayMs, cancellationToken);
                 balance2 = await lookupApi.BalancesAsync(include_all: false, limit: 1000, next: balance2.NextToken, round: round2, currency_greater_than: configuration.MinimumBalanceForsentasset, currency_less_than: null, asset_id: (int)configuration.assetsend, cancellationToken);

                balances2.AddRange(balance2.Balances);
            }
            return balances2;
        }

        public IEnumerable<Algorand.SignedTransaction> PrepareBatch(IEnumerable<KeyValuePair<string, ulong>> rewards, TransactionParametersResponse algoParams)
        {

            var ret = new List<Algorand.SignedTransaction>();
            var dispenserAccount = new Algorand.Account(configuration.DispenserMnemonic);
            var txsToSign = new List<Algorand.Transaction>();
            foreach (var rewardItem in rewards)
            {
                var receiverAddress = new Algorand.Address(rewardItem.Key);
                txsToSign.Add(Algorand.Utils.GetTransferAssetTransaction(dispenserAccount.Address, receiverAddress, assetId: configuration.assetsend, amount: rewardItem.Value, algoParams));
            }

            Algorand.Digest gid = Algorand.TxGroup.ComputeGroupID(txsToSign.ToArray());
            var signedTransactions = new List<Algorand.SignedTransaction>();
            foreach (var tx in txsToSign)
            {
                tx.AssignGroupID(gid);
                signedTransactions.Add(dispenserAccount.SignTransaction(tx));
            }
            return signedTransactions;
        }

        public Dictionary<string, ulong> CalculateAccountReward(IEnumerable<MiniAssetHolding> balances)
        {
            var ret = new Dictionary<string, ulong>();
            var interest = GetInterestPerInterval();
            foreach (var balance in balances)
            {
                var effectiveBalance = balance.Amount;
                if (effectiveBalance > configuration.MaximumBalanceForStaking)
                {
                    effectiveBalance = configuration.MaximumBalanceForStaking;
                }
                ret[balance.Address] = Convert.ToUInt64(Math.Round(Convert.ToDecimal(effectiveBalance) * interest));
            }
            return ret;
        }

        /// <summary>
        /// Returns current interest per interval from current configuration
        /// </summary>
        /// <returns></returns>
        public decimal GetInterestPerInterval()
        {
            var annualInterest = configuration.InterestRate / 100;  // 10 > 0,1
            var intervalInSeconds = configuration.Interval;         // 3600
            var secondsPerYear = (ulong)31536000;                  // 31536000
            var intervalsPerYear = Convert.ToInt32(secondsPerYear / intervalInSeconds); // 8 760
            var powerBase = Convert.ToDouble(annualInterest + 1);
            var oneOverInterval = 1 / Convert.ToDouble(intervalsPerYear);
            var powered = Math.Pow(powerBase, oneOverInterval);
            var interestPerInterval = Convert.ToDecimal(powered - 1);
            return interestPerInterval;
        }
    }
}